/*
 * Copyright 2016-2017 Golden Code Development Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"),  to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Export all public tables to <_file._dump-name>.d files in an export directory off current
 * working directory.
 *
 * Simplifying assumptions: no raw, clob, blob, datetime, or datetime-tz fields; metadata tables
 * not dumped.
 *
 * There are two known problems with the file footer written out in each dump file:
 * 1) the session:cpstream attribute is not properly implemented at this time and always returns
 *    "UTF-8";
 * 2) the CRC is bogus.
 */

define stream dump-stream.

define variable hQry    as handle no-undo.
define variable hBuf    as handle no-undo.
define variable hFld    as handle no-undo.
define variable xDir    as char init "./export" no-undo.
define variable delim   as char init " " no-undo.
define variable line    as char no-undo.
define variable ldb     as char no-undo.
define variable recs    as int no-undo.
define variable num     as int no-undo.
define variable ext     as int no-undo.
define variable idx     as int no-undo.
define variable dtnow   as datetime no-undo.

/*
 * Purpose:    Process a single scalar or extent value from a buffer field.
 * Parameters: hFld
 *             Handle to the buffer field.
 *             idx
 *             1-based index of the field value for an extent field; 0 for a scalar field.
 *             delim
 *             Delimiter character.
 *
 * Returns:    Value, formatted for export, followed by delimiter character.
 */
function process-value returns char (input hFld as handle, input idx as int, input delim as char):

   define variable result as char no-undo.
   
   if idx = 0 then
   /* scalar field */
   do:
      if hFld:buffer-value() = ? then
      do:
         result = "?".
      end.
      else do:
         case hFld:data-type:
            when "character" then result = quoter(hFld:buffer-value()).
            otherwise result = string(hFld:buffer-value()).
         end.
      end.
   end.
   /* extent field */
   else do:
      if hFld:buffer-value(idx) = ? then
      do:
         result = "?".
      end.
      else do:
         case hFld:data-type:
            when "character" then result = quoter(hFld:buffer-value(idx)).
            otherwise result = string(hFld:buffer-value(idx)).
         end.
      end.
   end.
   
   return result + delim.
   
end.

/* Main Block */

/* delete any previous version of the export directory, then create a fresh one */
os-delete value(xDir) recursive.
os-create-dir value(xDir).

/* loop through all public tables and dump each one to a .d file */
for each _file where _file._owner = "PUB" and not _file._file-name begins "_" no-lock:

   output stream dump-stream to value(xDir + "/" + _file._dump-name + ".d").
   
   create buffer hBuf for table _file._file-name.
   create query hQry.
   
   ldb = hBuf:dbname.
   
   hQry:set-buffers(hBuf).
   hQry:query-prepare("for each " + _file._file-name).
   hQry:query-open().
   hQry:get-first().
   
   recs = 0.
   
   do while not hQry:query-off-end:
   
      line = "".
      
      do num = 1 to hBuf:num-fields:
      
         hFld = hBuf:buffer-field(num).
         ext = hFld:extent.
         if ext = 0 then
         do:
            line = line + process-value(hFld, 0, delim).
         end.
         else do idx = 1 to ext:
            line = line + process-value(hFld, idx, delim).
         end.
      
      end.
      
      /* right-trim result */
      line = right-trim(line, delim).
      
      put stream dump-stream unformatted line skip.
      
      recs = recs + 1.
      
      hQry:get-next().
      
   end.
   
   /* write export file footer */
   dtnow = now.
   put stream dump-stream unformatted
      "." skip
      "PSC" skip
      "filename=" + _file._file-name skip
      "records=" + string(recs, "9999999999999") skip
      "ldbname=" + ldb skip
      "timestamp=" + string(year(dtnow), "9999") + "/"
                   + string(month(dtnow), "99") + "/"
                   + string(day(dtnow), "99")
                   + "-" + string(integer(mtime(dtnow) / 1000), "hh:mm:ss") skip
      "numformat=" + string(asc(session:numeric-separator))
                   + "," + string(asc(session:numeric-decimal-point)) skip
      "dateformat=" + session:date-format + "-" + string(session:year-offset) skip
      "map=NO-MAP" skip
      "cpstream=" + session:cpstream skip
      "." skip
      "0000000000" skip.
   
   /* close stream */
   output stream dump-stream close.
   
   hQry:query-close().
   
end.

quit.
