/**
 * API for the "heat map" calendar control.
 *
 * Displays a horizontal calendar, centered around a specific date. Displays the month containing
 * that date and as many months as can fit to the left and right within a specified pixel width,
 * assuming a fixed cell size for each date. Each week is represented by a column; thus, the
 * core calendar is seven rows high (one per day of week).
 *
 * The calendar is initialized with a call to the initialize function and rendered (and updated)
 * with the render function. The former sets the initial date and maximum width of the calendar.
 * The latter associates date with the dates and renders. The initialize function should be
 * invoked only once. The render function can be invoked multiple times.
 *
 * Each date cell displays a color indicating the "heat" of the data associated with that date.
 * The "heat" is represented by a score within a finite range. The value of the score determines
 * which fill color is used to represent the associated date. The available fill colors are
 * encoded in a series of CSS classes, and one of these is applied based on the score value.
 * The CSS classes should be defined in such a way as to signify a progression from "cool" to
 * "hot".
 *
 * A title indicating the date and associated score is associated with each date cell in the
 * calendar.
 *
 * The user can interact with the calendar by clicking on a date, or by swiping a series of
 * contiguous dates. A callback function is invoked upon one of these actions, allowing a caller
 * to perform arbitrary functionality based on the date range selected.
 *
 * Data for a contiguous range of dates can be updated with the render function, which will update
 * the fill color and titles of those dates. This event is depicted by changing the fill color of
 * all affected cells temporarily to blue, then transitioning them to their new fill color.
 *
 * The initial date upon which the calendar is centered is represented visually by a blue circle
 * overlaid atop the corresponding date cell.
 *
 * Derived from a sample by Joshua Golub
 * (https://bl.ocks.org/joshua-golub/2f7a5cc3be8f1f3a170bacdeebb987b5/eb68f534fde5d07c424e26a88767d0efed454e81),
 * which in turn was based on works by Micah Stubbs
 * (https://bl.ocks.org/micahstubbs/89c6bd879d64aa511372064c6cf85711) and Michael Bostock
 * (https://bl.ocks.org/mbostock/4063318).
 *
 * This work contains significant changes by Golden Code Development.
 *
 * Released under the GNU General Public License, version 3
 * (https://opensource.org/licenses/GPL-3.0).
 */
define(function()
{
   /** Object which contains API functions. */
   var me = {};
   
   /** First date represented in calendar. */
   var startDate;
   
   /** Last date represented in calendar. */
   var endDate;
   
   /** Height and width of each date rectangle in calendar. */
   var cellSize;
   
   /** Width available to calendar SVG */
   var svgWidth;
   
   /** Number of full or partial weeks represented in calendar */
   var totalWeeks;
   
   /** First date in calendar which represents a Sunday */
   var firstSunday;
   
   /** D3 selection of SVG rect elements representing dates */
   var dayRects;
   
   /**
    * The `d3Selection` type represents a D3 selection object. Learn more about D3 selections
    * [here](https://github.com/d3/d3/blob/master/API.md#selections-d3-selection).
    * 
    * @typedef {object} d3Selection
    */
   
   /**
    * The `d3Map` type represents a D3 map object. Learn more about D3 maps
    * [here](https://github.com/d3/d3-collection/blob/master/README.md#maps).
    * 
    * @typedef {object} d3Map
    */

   /**
    * Generate an <svg> element for the calendar, containing a single <g> element.
    *
    * @param   {DOM element} parentDiv
    *          DOM element to which the SVG will be appended.
    * @param   {number} height
    *          The height to use for the <svg>
    *
    * @return  A D3 selection of the generated <svg> element representing the calendar.
    */
   function generateCalendar(parentDiv, height)
   {
      /**
       * Return the X-axis offset to use for the <g> element representing a given year.
       *
       * @return {!number} The X-axis offset
       */
      function getOffsetX()
      {
         return (svgWidth - cellSize * totalWeeks) / 2;
      }

      /**
       * Return the X-axis offset to use for the <g> element representing a given year.
       *
       * @return {!number} The X-axis offset
       */
      function getOffsetY()
      {
         return (height - cellSize * 9);
      }

      return d3.select(parentDiv)
               .append("svg")
               .attr("width", svgWidth)
               .attr("height", height)
               .attr("class", "year")
               .append("g")
               .attr("transform", "translate(" + getOffsetX() + "," + getOffsetY() + ")");
   }
   
   /**
    * Append two gradient definitions to the calendar SVG:
    * 1) a radial gradient used when rendering the user-selected date range path;
    * 2) a linear gradient used to render a color map legend.
    * 
    * Render a rectangle that serves as a legend to the calendar heat map, using the second of
    * these. Inspect the color of each style sheet associated with the date rectangle colors and
    * use that information to define the linear gradient.
    * 
    * @param   {string} lowText
    *          Text to render at the low end of the color map gradient legend.
    * @param   {string} highText
    *          Text to render at the high end of the color map gradient legend.
    */
   function appendGradientDef(calendar, lowText, highText)
   {
      var defs = calendar.append("defs");
      
      // set up gradient for date range selection path
      var grad = defs.append("radialGradient")
                     .attr("id", "dateRangeGradient")
                     .attr("fx", "25%")
                     .attr("fy", "25%")
                     .attr("r", "65%")
                     .attr("spreadMethod", "pad");
      grad.append("stop")
          .attr("offset", "0%")
          .attr("stop-color", "#fff")
          .attr("stop-opacity", 1);
      grad.append("stop")
          .attr("offset", "100%")
          .attr("stop-color", "#850")
          .attr("stop-opacity", 1);
      
      // set up gradient for color map
      grad = defs.append("linearGradient")
                 .attr("id", "colorMapGradient")
                 .attr("x1", "0%")
                 .attr("y1", "0%")
                 .attr("x2", "100%")
                 .attr("y2", "0%")
                 .attr("spreadMethod", "pad");
      var sheets = document.styleSheets;
      for (var i = 0; i < sheets.length; i++)
      {
         var classes = sheets[i].rules ||
                       sheets[i].cssRules ||
                       sheets[i].rules[0] ||
                       sheets[i].rules[0].cssRules;
         var count = 0;
         var colors = [];
         for (var j = 0; j < classes.length; j++)
         {
            var cls = classes[j];
            if (cls.selectorText && cls.selectorText.startsWith(".year .color"))
            {
               colors[count++] = cls.style.fill;
            }
         }
      }
      
      var offset = 110 / (count + 1);
      for (var i = count - 1; i >= 0; i--)
      {
         grad.append("stop")
             .attr("offset", (105 - (offset * (i + 1))) + "%")
             .attr("stop-color", colors[i])
             .attr("stop-opacity", "1");
      }
      
      var cmWidth = cellSize * count;
      var cmX = cellSize * (totalWeeks - count);
      var textY = cellSize * 8;
      
      // render the color map legend
      calendar.append("rect")
              .attr("x", cmX)
              .attr("y", cellSize * 7 + 4)
              .attr("width", cellSize * count)
              .attr("height", cellSize)
              .attr("fill", 'url("#colorMapGradient")')
              .attr("stroke", "black");
      
      // apply labels to the color map legend
      calendar.append("text")
              .attr("class", "color-map-label")
              .attr("transform", "translate(" + (cmX + 2) + ", " + textY + ")")
              .text(lowText);
      calendar.append("text")
              .attr("class", "color-map-label")
              .attr("transform", "translate(" + (cmX + cmWidth - 2) + ", " + textY + ")")
              .style("text-anchor", "end")
              .attr("fill", "white")
              .text(highText);
   }
   
   /**
    * Attach <text> elements for the starting and ending years, day of week labels, plus a help
    * message.
    *
    * @param   {!d3Selection} calendar
    *          A D3 selection object representing the <g> element of the calendar.
    */
   function attachTextToCalendar(calendar)
   {
      // affix year of start date to left of calendar
      calendar.append("text")
              .attr("transform", "translate(" + (-cellSize - 4) +"," + cellSize * 3.5 + ")rotate(-90)")
              .style("text-anchor", "middle")
              .text(startDate.getFullYear());
      
      // affix year of end date to right of calendar
      calendar.append("text")
              .attr("transform", "translate(" + ((cellSize * (totalWeeks + 1)) + 4) + "," + cellSize * 3.5 + ")rotate(90)")
              .style("text-anchor", "middle")
              .text(endDate.getFullYear());
      
      // affix day of week labels to left of calendar
      var days = "SMTWTFS";
      for (var i = 0; i < 7; i++)
      {
         calendar.append("text")
                 .attr("transform", "translate(-3," + ((cellSize * (i + 1)) - 4) + ")")
                 .attr("class", "day-label")
                 .text(days.charAt(i));
      }
      
      // affix help message below calendar on left
      calendar.append("text")
              .attr("id", "dateRangeLabel")
              .attr("transform", "translate(0," + cellSize * 8 + ")")
              .text("Click or swipe to select dates...");
   }
   
   /**
    * Return the week offset for the given date, where 0 represents the starting week of the
    * calendar, which is not necessarily the first of the year.
    *
    * @param   {!Date} date
    *          Date for which to calculate the week offset.
    *
    * @return  {number}
    *          The week offset number.
    */        
   function getWeekOffset(date)
   {
      return moment(date).diff(firstSunday, "weeks");
   }
   
   /**
    * Return the day number for the given date, where 0 represents Sunday and
    * 6 represents Saturday.
    *
    * @param   {!Date} date
    *          The date.
    *
    * @return  {number} The day number
    */
   function getDayNumber(date)
   {
      return date.getDay();
   }

   /**
    * Perform the following actions:
    *
    *     1. Generate one <rect class="day"> for each day of each year
    *     2. Attach each generated <rect> to the <g class="year"> for the associated year
    *     3. Attach the date object for the given day as the data for the <rect>
    *     4. Convert the date objects associated with each <rect> to a date string
    *
    * @param   {!d3Selection} calendar
    *          A D3 selection object representing the <g> element of the calendar.
    *
    * @return  {d3Selection} A D3 selection representing the generated rectangles
    */
   function generateDayRects(calendar)
   {
      /**
       * Return the X-axis coordinate to use for the top-left corner of the rectangle
       * representing the given date.
       *
       * @param   {!Date} date
       *          The date
       *
       * @return  {number} The X-axis coordinate.
       */
      function getX(date)
      {
         return getWeekOffset(date) * cellSize;
      }

      /**
       * Return the Y-axis coordinate to use for the top-left corner of the rectangle
       * representing the given date.
       *
       * @param   {!Date} date
       *          The date.
       *
       * @return  {number} The Y-axis coordinate.
       */
      function getY(date)
      {
         return getDayNumber(date) * cellSize;
      }

      /**
       * Generate and return an array of date objects representing all of the days in the
       * calendar.
       *
       * @return {Array.<Date>} The array of date objects.
       */
      function getAllDaysInCalendar()
      {
         return d3.timeDays(startDate, endDate);
      }

      return calendar.selectAll(".day")
                     .data(getAllDaysInCalendar)      
                     .enter()
                     .append("rect")
                     .attr("class", "day color0")
                     .attr("width", cellSize)
                     .attr("height", cellSize)
                     .attr("x", getX)
                     .attr("y", getY);
   }
    
   /**
    * Attach a <title> element to each of the given day rectangles, using the same data string
    * that was generated for the rectangle. Register event handlers for mouse up, down, and move
    * events to handle user clicking/swiping.
    *
    * @param   {!d3Selection} dayRects
    *          A D3 selection object of the date rectangles in the calendar.
    * @param   {dateRangeFunc} dateRangeFunc
    *          A callback to be invoked when the user clicks on a date in the calendar, or swipes
    *          the mouse to select multiple dates.
    */
   function finishDayRectSetup(dayRects, dateRangeFunc)
   {
      var start, end;
      
      /**
       * Function invoked on mousedown event. Starts a click or swipe action.
       */
      function swipeStart()
      {
         var date = d3.select(this).datum();
         start = date;
         end = date;
         updatePathAndLabel(start);
      }
      
      /**
       * Function invoked on mousemove event. Updates date selection range if currently involved
       * in a swipe action.
       */
      function swiped()
      {
         var date = d3.select(this).datum();
         if (!start)
         {
            end = undefined;
            return;
         }
         if (end == date)
         {
            return;
         }
         end = date;
         updatePathAndLabel(end);
      }
      
      /**
       * Function invoked on mouseup event. Completes a click or swipe action and invokes the
       * date range callback function provided by code which rendered the calendar. 
       */
      function swipeEnd()
      {
         if (!start || !end)
         {
            return;
         }
         var date = d3.select(this).datum();
         var a = updatePathAndLabel(date);
         start = undefined;
         end = undefined;
         dateRangeFunc(a[0], a[1]);
      }
      
      /**
       * Invoked during mouse up, down, and move events to render the path the user is selecting
       * or has selected.
       *
       * @param   {date} end
       *          End date of clicked/swiped range. Note that this can be earlier than the start
       *          date if the user swiped right to left, but this is normalized by the function.
       */
      function updatePathAndLabel(end)
      {
         var a = (start <= end) ? [ start, end ] : [ end, start ];
         var path = getPathDescriptionForRange(a[0], a[1]);
         d3.select("#targetRange")
           .attr("d", path);
         var label;
         if (end == start)
         {
            label = d3.timeFormat("%Y-%m-%d")(start);
         }
         else
         {
            label = d3.timeFormat("%Y-%m-%d")(a[0]) +
            " thru " +
            d3.timeFormat("%Y-%m-%d")(a[1]);
         }
         d3.select("#dateRangeLabel")
           .text(label);
         
         return a;
      }
      
      dayRects.on("mousedown", swipeStart)
              .on("mousemove", swiped)
              .on("mouseup", swipeEnd)
              .append("title")
              .text(function(d) { return d3.timeFormat("%Y-%m-%d")(d); });
    }
    
   /**
    * Perform the following actions:
    *
    *     1. Generate one <path class="month"> for each month of each year
    *     2. Attach each generated <path> to the <g class="year"> for the associated year
    *     3. Attach the date object for the given month as the data for the <p>
    *     4. Attach the path description for the given month as the @d attribute of the <p>
    * 
    * @param   {!d3Selection} calendar
    *          A D3 selection object representing the <g> element for the calendar control.
    */
   function appendMonthPathsToCalendar(calendar)
   {
      /**
       * Generate and return an array of date objects representing all of the months in the
       * calendar.
       *
       * @return {Array.<Date>} The array of date objects
       */
      function getAllMonths()
      {
         return d3.timeMonths(startDate, endDate);
      }
      
      /**
       * Generate and return the <path> description for the given month, where the path
       * is the outline drawn around each month in the following representation of a year:
       * 
       *      JAN   FEB  MAR  APR   MAY  JUN  JUL   AUG  SEP   OCT  NOV  DEC 
       *       |     |    |    |     |    |    |     |    |     |    |    |
       *       
       *      -----------------------------------------------------------------
       *      |1111|2222|3333|44444|5555|6666|77777|8888|99999|0000|AAAA|BBBBB|  --- Sundays
       *     |11111|2222|3333|44444|5555|6666|77777|8888|9999|00000|AAAA|BBBBB|  --- Mondays
       *     |11111|2222|3333|4444|55555|6666|77777|8888|9999|00000|AAAA|BBBB|   --- Tuesdays
       *     |11111|2222|3333|4444|55555|6666|7777|88888|9999|00000|AAAA|BBBB|   --- Wednesdays
       *     |1111|2222|33333|4444|55555|6666|7777|88888|9999|0000|AAAAA|BBBB|   --- Thursdays
       *     |1111|2222|33333|4444|5555|66666|7777|88888|9999|0000|AAAAA|BBBB|   --- Fridays
       *     |1111|2222|33333|4444|5555|66666|7777|8888|99999|0000|AAAA|BBBBB|   --- Saturdays
       *     -----------------------------------------------------------------
       * 
       * Breaking out the outline, the general format looks like this:
       *    
       *      7----6
       *      |nnnn|
       *      |nnnn|
       *      |nnnn|
       *     10nnnn|
       *     |nnnnn|
       *     |nnnn45
       *     |nnnn|
       *     2----3
       *  
       *  where 0-7 indicate the vertices of the shape, in the order in which the outline
       *  will be drawn.
       *
       * TODO: refactor to use getPathDescriptionForRange instead of duplicating much of this
       * logic there.
       *    
       * @param   {!Date} month
       *          The date object representing the month.
       *
       * @return  {string} The path description
       */
      function getPathDescriptionForMonth(month)
      {
         var nextMonth,                //  date object for the first of the month after the given one
             nextMonthDayOfWeek,       //  0 (sunday) - 6 (saturday)
             nextMonthWeekOffset,      //  first of next month's distance in weeks from starting sunday
             path,
             thisMonthDayOfWeek,       //  0 (sunday) - 6 (saturday)
             thisMonthWeekOffset;      //  first of this month's distance in weeks from starting sunday
          
          nextMonth = moment(month).add(1, "months").toDate();
          thisMonthDayOfWeek = month.getDay();
          thisMonthWeekOffset = getWeekOffset(month);
          nextMonthDayOfWeek = nextMonth.getDay();
          nextMonthWeekOffset = getWeekOffset(nextMonth);

          //  generate a path of the form:
          //  
          //      M{0.x},{0.y}H{1.x}V{2.y}H{3.x}V{4.y}H{5.x}V{6.y}H{7.x}
          //
          //  where:
          //  
          //      {n.x} represents the x coordinate of point `n` in the per-month chart, above
          //      {n.y} represents the y coordinate of point `n` in the per-month chart, above 
          //
          path = "M" + (thisMonthWeekOffset + 1) * cellSize + "," + thisMonthDayOfWeek * cellSize  //  moveto #0
               + "H" + thisMonthWeekOffset * cellSize                                              //  lineto #1
               + "V" + 7 * cellSize                                                                //  lineto #2
               + "H" + nextMonthWeekOffset * cellSize                                              //  lineto #3
               + "V" + nextMonthDayOfWeek * cellSize                                               //  lineto #4
               + "H" + (nextMonthWeekOffset + 1) * cellSize                                        //  lineto #5
               + "V" + 0                                                                           //  lineto #6
               + "H" + (thisMonthWeekOffset + 1) * cellSize                                        //  lineto #7
               + "Z";                                                                              //  close path

          return path;                                                                            
      }
      
      /**
       * Determine the translation that should be applied to the month label for the given month,
       * based on its position in the calendar.
       *
       * @param   {date} month.
       *          First date of the given month.
       *
       * @return  Translate string to be applied to month name text element.
       */
      function getTranslateForMonthLabel(month)
      {
         var nextMonth = moment(month).add(1, "months").toDate();
         var thisMonthWeekOffset = getWeekOffset(month);
         var nextMonthWeekOffset = getWeekOffset(nextMonth);
         
         return "translate(" + ((nextMonthWeekOffset + thisMonthWeekOffset + 1) * cellSize / 2) + ", -3)";
      }
      
      // append month paths
      calendar.selectAll(".month")
              .data(getAllMonths)
              .enter()
              .append("path")
              .attr("class", "month")
              .attr("d", getPathDescriptionForMonth);
      
      // append month names; these must be appended to the month path parent node
      calendar.selectAll(".month-label")
              .data(getAllMonths)
              .enter()
              .append("text")
              .text(d3.timeFormat("%b"))
              .datum(getTranslateForMonthLabel)
              .attr("class", "month-label")
              .attr("transform", function(d) { return d; });
   }
   
   /**
    * Compose an SVG path description for an arbitrary, contguous block of date rectangles in
    * the calendar.
    *
    * @param   {date} startDate
    *          Starting date of range.
    * @param   {date} endDate
    *          Ending date of range.
    *
    * @return  SVG path description which includes all date rectangles in the given range.
    */
   function getPathDescriptionForRange(startDate, endDate)
   {
      var startWeekOffset = getWeekOffset(startDate);
      var startDayOfWeek = startDate.getDay();
      var endWeekOffset = getWeekOffset(endDate);
      var endDayOfWeek = endDate.getDay();
      var bottom = 7, top = 0;
      if (startWeekOffset == endWeekOffset)
      {
         bottom = endDayOfWeek + 1;
         top = startDayOfWeek;
      }
      
      var path = "M" + (startWeekOffset + 1) * cellSize + "," + startDayOfWeek * cellSize  //  moveto #0
               + "H" + startWeekOffset * cellSize                                          //  lineto #1
               + "V" + bottom * cellSize                                                   //  lineto #2
               + "H" + endWeekOffset * cellSize                                            //  lineto #3
               + "V" + (endDayOfWeek + 1) * cellSize                                       //  lineto #4
               + "H" + (endWeekOffset + 1) * cellSize                                      //  lineto #5
               + "V" + top * cellSize                                                      //  lineto #6
               + "H" + (startWeekOffset + 1) * cellSize                                    //  lineto #7
               + "Z";                                                                      //  close path
      
      return path;                                                                            
   }

   /**
    * Update the class attribute and <title> element for each day <rect> that appears in the
    * given data set.
    *
    * @param   {!d3Selection} dayRects
    *          A d3 selection of <rect> elements associated with calendar dates.
    * @param   {object} data
    *          JSON data object containing data to associate with date rectangles.
    * @param   {boolean} isUpdate
    *          True if this call represents an update to an existing calendar; false if calendar
    *          is being rendered for the first time.
    */
   function updateClassAndTitleForDayRects(dayRects, jsonData, isUpdate)
   {
      var color,
          dataMap,
          percentFormatter;
          
      /**
       * Generate a return a D3 scale function for mapping values from -0.05 to +0.05
       * to one of eleven colors. See [here](https://github.com/d3/d3-scale/blob/master/README.md#scaleQuantize)
       * for more information about quantize scales.
       *
       * @return {function} The scale function.
       */
      function getColorScale()
      {
         return d3.scaleQuantize()
                  .domain(jsonData.domain)
                  .range(d3.range(11).map(function(int) { return "color" + int; }));
      }

      /**
       * Return a boolean indication of whether the given date exists in the data set.
       *
       * @param   {!date} date
       *          The date.
       *
       * @return  {boolean} True if the given date exists in the data set; false otherwise.
       */
      function isRectDataInDataMap(date)
      {
         return dataMap.has(date);
      }

      /**
       * Generate and return the value of the class attribute to use for a <rect> having the given
       * date as its data.
       *
       * @param   {!date} date
       *          The date.
       *
       * @return  {string} The class attribute value.
       */
      function getRectClass(date)
      {
         return "day " + color(dataMap.get(date));
      }

      /**
       * Generate and return the title to use for a <rect> having the given date as its data.
       *
       * @param   {!string} dateString
       *          The date.
       *
       * @return  {string} The title string.
       */
      function getRectTitle(date)
      {
         return d3.timeFormat("%Y-%m-%d")(date) + ": " + percentFormatter(dataMap.get(date));
      }
      
      /**
       * Get a date lookup key for an item from that item's date string.
       *
       * @param   {object} item
       *          Item associated with day rectangle.
       *
       * @return  Date associated with item.
       */
      function getDataKey(item)
      {
         var a = item.date.split(/[^0-9]/);
         var d = new Date(a[2], a[0] - 1, a[1]);
         
         return d;
      }
      
      /**
       * Get the score value from a data item.
       *
       * @param   {object} item
       *          Item associated with day rectangle.
       */
      function getDataValue(items)
      {
         return items[0].score;
      }
      
      dataMap = d3.nest()
                  .key(getDataKey)
                  .rollup(getDataValue)
                  .map(jsonData.rows);
      color = getColorScale();
      percentFormatter = d3.format(".1%");
      
      dayRects.filter(isRectDataInDataMap)
              .attr("class", getRectClass)
              .select("title")
              .text(getRectTitle);
      
      // if this call represents an update to an existing calendar (as opposed to rendering the
      // calendar for the first time), indicate that the affected block of dates has been
      // updated by setting the fill color for the associated date rectangles to blue, then
      // gradually fading back to the original fill color
      if (isUpdate)
      {
         dayRects.filter(isRectDataInDataMap)
                 .transition().style("fill", "#1589ff").duration(250)
                 .transition().style("fill", null).duration(6000);
      }
   }
   
   /**
    * Mark the current date by appending a blue circle over the date rectangle representing
    * today's date.
    *
    * @param   {!d3Selection} calendar
    *          A D3 selection object representing the <g> element for the calendar control.
    */
   function markToday(calendar)
   {
      var today = new Date();
      var cx = (getWeekOffset(today) + 0.5) * cellSize;
      var cy = (getDayNumber(today) + 0.5) * cellSize;
      
      calendar.append("circle")
              .attr("cx", cx)
              .attr("cy", cy)
              .attr("r", cellSize / 4)
              .attr("fill", "#2982CC")
              .attr("pointer-events", "none");
   }
   
   /**
    * Append a path to the calendar SVG to signify a date range the user has chosen. The path is
    * updated when the user clicks or swipes dates, conforming to a contiguous block of the
    * selected dates.
    *
    * @param   {d3Selection} calendar
    *          A D3 selection object representing the <g> element for the calendar control.
    */
   function appendTargetRangePath(calendar)
   {
      calendar.append("path")
              .attr("id", "targetRange")
              .attr("class", "target-range")
              .attr("pointer-events", "none");
   }

   /**
    * Render the data in the given JSON object within the calendar, either for the first time,
    * or as an update to an existing calendar.
    *
    * @param   {object} jsonData
    *          Date data to be rendered by the calendar.
    * @param   {DOM element} parentDiv
    *          Parent DOM element to which to attach the calendar (only used if the calendar does
    *          not already exist).
    * @param   {function} dateRangeFunc
    *          Callback function to be invoked when the user clicks a date or swipes a range of
    *          dates in the calendar (only used if the calendar does not already exist).
    * @param   {string} lowText
    *          Text to render at the low end of the color map gradient legend.
    * @param   {string} highText
    *          Text to render at the high end of the color map gradient legend.
    */
   me.render = function render(jsonData, parentDiv, dateRangeFunc, lowText, highText)
   {
      if (dayRects)
      {
         updateClassAndTitleForDayRects(dayRects, jsonData, true);
      }
      else
      {
         var height,
             calendar;

         height = cellSize * 10;
         calendar = generateCalendar(parentDiv, height);
         appendGradientDef(calendar, lowText, highText);
         attachTextToCalendar(calendar);
         dayRects = generateDayRects(calendar);
         finishDayRectSetup(dayRects, dateRangeFunc);
         appendMonthPathsToCalendar(calendar);
         updateClassAndTitleForDayRects(dayRects, jsonData);
         markToday(calendar);
         appendTargetRangePath(calendar);
      }
   }
   
   /**
    * Initialize the calendar control, based on the given, initial date and the available width.
    * The calendar is roughly centered on the initial date, and contains as many months to the
    * left and right of that date's month as can fit in the given width.
    *
    * This function does not render the control. It only calculates some key values that will be
    * necessary at render time and returns the start and end dates it has determined for the
    * calendar. The start date will always be the first date of the first month displayed by the
    * calendar. Then end date will always be the last date of the last month displayed by the
    * calendar.
    *
    * Current limitation:
    * It is assumed that there is at least enough width available to fit one month, along with
    * the surrounding text. No error checking is done for scenarios where less width is actually
    * available.
    *
    * @param   {date} initialDate
    *          Initial date on which the calendar should be centered.
    * @param   {int} width
    *          Width, in pixels, available to the calendar control.
    *
    * @return  Array containing the start and end dates calculated for the calendar at the given
    *          width.
    */
   me.initialize = function initialize(initialDate, width)
   {
      // hard coded cell size
      cellSize = 17;
      
      // remember the available width
      svgWidth = width;
      
      // calculate the maximum number of week columns we could fit in the allotted width
      var maxCols = (width - (cellSize * 3)) / cellSize;
      
      // determine how many weeks to pad on each side of our initial date, not yet taking month
      // boundaries into account
      var weeksPast = Math.floor(maxCols / 2) - 1;
      var weeksFuture = Math.ceil(maxCols / 2);
      
      // determine the date range if we place initialDate roughly at the center of that interval
      var startMoment = moment(initialDate).subtract(weeksPast, "weeks");
      var endMoment = moment(initialDate).add(weeksFuture, "weeks");
      
      // if start is not first of month, set forward to beginning of next month
      if (startMoment.date() > 1)
      {
         startMoment.month(startMoment.month() + 1).date(1);
      }
      
      // if end is not last of month, set backward to end of previous month
      var dateOfMonth = endMoment.date();
      if (dateOfMonth < endMoment.daysInMonth())
      {
         endMoment.date(1).subtract(1);
      }
      
      // calculate some key values that will be needed to render the control
      startDate = d3.timeDay.floor(startMoment.toDate());
      endDate = d3.timeDay.floor(endMoment.toDate());
      firstSunday = moment(startDate).day("Sunday");
      totalWeeks = endMoment.diff(firstSunday, "weeks");
      
      return [ startDate, endDate ];
   }
   
   return me;
});
